from io import StringIO
from django.core.management import call_command
from django.test import TestCase


class CheckViewPermsTest(TestCase):

    def test_missing_perms(self):
        out = StringIO()
        call_command('check_view_permissions', stdout=out)
        self.assertIn('`tests.test_perm`', out.getvalue())
        self.assertIn('`tests.test_perm2`', out.getvalue())



"""Views used for testing."""
from django.contrib.auth.mixins import (
    LoginRequiredMixin, PermissionRequiredMixin, UserPassesTestMixin
)
from django.views.generic import View


class LoginRequiredView(LoginRequiredMixin, View):
    pass


class PermissionRequiredView(PermissionRequiredMixin, View):
    permission_required = 'tests.test_perm'


class PermissionRequiredMultiView(PermissionRequiredMixin, View):
    permission_required = ('tests.test_perm', 'tests.test_perm2')


class PermissionRequiredViewNoPerm(PermissionRequiredMixin, View):

    def has_permission(self):
        """The user's first name must be Bob"""
        return self.request.user.first_name == 'Bob'


class PermissionRequiredViewDocstring(PermissionRequiredMixin, View):
    permission_required = 'tests.test_perm'

    def has_permission(self):
        """Custom docstrings should be detected."""
        return super().has_permission()


class PermissionRequiredViewNoDocstring(PermissionRequiredMixin, View):
    permission_required = 'tests.test_perm'

    def has_permission(self):
        return super().has_permission()


class UserPassesTestView(UserPassesTestMixin, View):
    def test_func(self):
        return True


class UserPassesTestViewDocstring(UserPassesTestMixin, View):
    def test_func(self):
        """Custom docstrings should be detected."""
        return True


class UserPassesTestViewNoDocstring(UserPassesTestMixin, View):
    def test_func(self):
        return True


class UserPassesTestViewCustomFunc(UserPassesTestMixin, View):
    def get_test_func(self):
        return self.custom_test_func

    def custom_test_func(self):
        """Custom docstrings should be detected."""
        return True



SECRET_KEY = 'test-key'
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
    }
}
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3'
    }
}
MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
]
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'permissions_auditor',
]
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
ROOT_URLCONF = 'permissions_auditor.tests.fixtures.urls'



from django.contrib.auth.models import Permission
from django.core.management.base import BaseCommand

from permissions_auditor.core import get_views


class Command(BaseCommand):
    help = 'Parse all view permissions, and find any that are missing from the database.'

    def get_view_permissions(self):
        views = get_views()
        permissions = []

        for view in views:
            permissions.extend(view.permissions)

        return list(set(permissions))

    def get_db_permissions(self):
        permissions = Permission.objects.all().values_list('content_type__app_label', 'codename')
        return ["{}.{}".format(ct, codename) for ct, codename in permissions]

    def handle(self, *args, **options):
        view_permissions = self.get_view_permissions()
        db_permissions = self.get_db_permissions()
        missing_perms = []

        for permission in view_permissions:
            if permission not in db_permissions:
                missing_perms.append(permission)

        if missing_perms:
            for permission in missing_perms:
                self.stdout.write(self.style.WARNING(
                    'Warning: No database entry found for permission `{}`.'.format(permission)
                ))
        else:
            self.stdout.write(self.style.SUCCESS('No permissions without database entries found.'))



import inspect

from django.conf import ImproperlyConfigured


class BaseProcessor:

    def can_process(self, view):
        """
        Can this processor process the provided view?

        :param view: the view being processed.
        :type view: function or class
        :return: whether this processor can process the view.
                 Default: ``False``
        :rtype: boolean
        """
        return True

    def get_permission_required(self, view):
        """
        Returns permissions required on the provided view.
        Must return an iterable.

        :param view: the view being processed.
        :type view: function or class
        :return: the permissions required to access the view. Default: ``[]``
        :rtype: list(str)
        """
        return []

    def get_login_required(self, view):
        """
        Returns if a user needs to be logged in to access the view.

        :param view: the view being processed.
        :type view: function or class
        :return: whether a user must be logged in to access this view.
                 Default: ``False``
        :rtype: boolean or None (if unknown)
        """
        return False

    def get_docstring(self, view):
        """
        Returns any additional information that should be displayed when
        showing permisison information.

        :param view: the view being processed.
        :type view: function or class
        :return: the string to display in the additional info column. Default: ``None``
        :rtype: str or None
        """
        return None


class BaseFuncViewProcessor(BaseProcessor):
    """Base class for processing function based views."""

    def can_process(self, view):
        return inspect.isfunction(view)


class BaseCBVProcessor(BaseProcessor):
    """Base class for processing class based views."""

    def can_process(self, view):
        return inspect.isclass(view)


class BaseDecoratorProcessor(BaseProcessor):
    """Base class with utilities for unwrapping decorators."""

    def _has_method_decorator(self, function, func_name):
        """
        Checks if a function with the name `func_name` (str) is present within the
        ``@method_decorator`` on the provided function.
        """
        closures = inspect.getclosurevars(function).nonlocals
        if 'decorators' in closures:
            for func in closures['decorators']:
                if func.__name__ == func_name or func.__qualname__.split('.')[0] == func_name:
                    return True

        if 'method' in closures:
            return self._has_method_decorator(closures['method'], func_name)

        return False

    def _get_method_decorators(self, function):
        """
        Returns a generator of functions that decorate the provided function using
        ``@method_decorator``.
        """
        closures = inspect.getclosurevars(function).nonlocals
        if 'decorators' in closures:
            for func in closures['decorators']:
                yield func

        if 'method' in closures:
            yield from self._get_method_decorators(closures['method'])

    def _has_test_func(self, function):
        """
        Checks if the provided function is decorated with the ``user_passes_test`` decorator.
        """
        closures = inspect.getclosurevars(function).nonlocals
        if 'test_func' in closures:
            return True

        if 'view_func' in closures:
            return self._has_test_func(closures['view_func'])

        return False

    def _has_test_func_lambda(self, function, name):
        """
        Checks if the provided function's test_func contains the lambda expression ``name`` (str).
        """
        closures = inspect.getclosurevars(function).nonlocals
        if 'test_func' in closures:
            if name in inspect.getclosurevars(closures['test_func']).unbound:
                return True

        if 'decorators' in closures:
            for func in closures['decorators']:
                if self._has_test_func_lambda(func, name):
                    return True

        if 'method' in closures:
            return self._has_test_func_lambda(closures['method'], name)

        return False

    def _get_test_func_closures(self, function):
        closures = inspect.getclosurevars(function).nonlocals
        if 'test_func' in closures:
            yield inspect.getclosurevars(closures['test_func'])

        if 'view_func' in closures:
            yield from self._get_test_func_closures(closures['view_func'])

    def _has_func_decorator(self, function, func_name):
        closures = inspect.getclosurevars(function).nonlocals
        if 'test_func' in closures:
            test_closures = inspect.getclosurevars(closures['test_func']).unbound
            if func_name in test_closures:
                return True

            if 'view_func' in closures:
                return self._has_func_decorator(closures['view_func'], func_name)

        return False


class BaseFileredMixinProcessor(BaseCBVProcessor):
    """
    Base class for parsing mixins on class based views.
    Set ``class_filter`` to filter the class names the processor applies to.
    ONLY checks top level base classes.

    :var class_filter: initial value: ``None``
    """
    class_filter = None

    def can_process(self, view):
        if not super().can_process(view):
            return False

        view_bases = [cls.__module__ + '.' + cls.__name__ for cls in view.__bases__]

        for cls_filter in self.get_class_filter():
            if cls_filter in view_bases:
                return True

        return False

    def get_class_filter(self):
        """
        Override this method to override the class_names attribute.
        Must return an iterable.

        :return: a list of strings containing the full paths of mixins to detect.
        :raises ImproperlyConfigured: if the ``class_filter`` atribute is ``None``.
        """
        if self.class_filter is None:
            raise ImproperlyConfigured(
                '{0} is missing the class_filter attribute. Define {0}.class_filter, or override '
                '{0}.get_class_filter().'.format(self.__class__.__name__)
            )
        if isinstance(self.class_filter, str):
            cls_filter = (self.class_filter, )
        else:
            cls_filter = self.class_filter
        return cls_filter
